import enum
import uuid
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from typing import Optional

import pytest
from serpyco_rs import Serializer

from .utils import repeat


@pytest.fixture(params=[True, False])
def validate(request) -> bool:
    return request.param


def test_dump_simple_types(benchmark):
    serializer = Serializer(int)
    benchmark.group = 'simple_types'
    benchmark(repeat(lambda: serializer.dump(1)))


def test_load_simple_types(benchmark, validate):
    serializer = Serializer(int)
    benchmark.group = 'simple_types'
    benchmark(repeat(lambda: serializer.load(1, validate=validate)))


def test_dump_optional(benchmark):
    serializer = Serializer(Optional[int])

    benchmark.group = 'optional'

    @benchmark
    def inner():
        repeat(lambda: serializer.dump(1))
        repeat(lambda: serializer.dump(None))


def test_load_optional(benchmark, validate):
    serializer = Serializer(Optional[int])

    benchmark.group = 'optional'

    @benchmark
    def inner():
        repeat(lambda: serializer.load(1, validate=validate))
        repeat(lambda: serializer.load(None, validate=validate))


def test_dump_list_simple_types(benchmark):
    serializer = Serializer(list[int])
    benchmark.group = 'list'
    data = list(range(1000))
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_list_simple_types(benchmark, validate):
    serializer = Serializer(list[int])
    benchmark.group = 'list'
    data = list(range(1000))
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


def test_dump_tuple_simple_types(benchmark):
    serializer = Serializer(tuple[int, str, bool])
    benchmark.group = 'tuple'
    benchmark(repeat(lambda: serializer.dump((123, 'foo', True))))


def test_load_tuple_simple_types(benchmark, validate):
    serializer = Serializer(tuple[int, str, bool])
    benchmark.group = 'tuple'
    benchmark(repeat(lambda: serializer.load((123, 'foo', True), validate=validate)))


def test_dump_dict_simple_types(benchmark):
    serializer = Serializer(dict[str, int])
    benchmark.group = 'dict'
    data = {str(i): i for i in range(1000)}
    benchmark(repeat(lambda: serializer.dump(data)))


@pytest.mark.slowtest
def test_load_dict_simple_types(benchmark, validate):
    serializer = Serializer(dict[str, int])
    benchmark.group = 'dict'
    data = {str(i): i for i in range(1000)}
    benchmark(repeat(lambda: serializer.load(data, validate=validate), count=100))


def test_dump_uuid(benchmark):
    serializer = Serializer(uuid.UUID)
    benchmark.group = 'uuid'
    data = uuid.uuid4()
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_uuid(benchmark, validate):
    serializer = Serializer(uuid.UUID)
    benchmark.group = 'uuid'
    data = str(uuid.uuid4())
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


def test_dump_date(benchmark):
    serializer = Serializer(date)
    benchmark.group = 'date'
    data = date.today()
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_date(benchmark, validate):
    serializer = Serializer(date)
    benchmark.group = 'date'
    data = date.today().isoformat()
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


def test_dump_time(benchmark):
    serializer = Serializer(time)
    benchmark.group = 'time'
    data = datetime.now().time()
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_time(benchmark, validate):
    serializer = Serializer(time)
    benchmark.group = 'time'
    data = datetime.now().time().isoformat()
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


def test_dump_datetime(benchmark):
    serializer = Serializer(datetime)
    benchmark.group = 'datetime'
    data = datetime.now()
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_datetime(benchmark, validate):
    serializer = Serializer(datetime)
    benchmark.group = 'datetime'
    data = datetime.now().isoformat()
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


def test_dump_decimal(benchmark):
    serializer = Serializer(Decimal)
    benchmark.group = 'decimal'
    data = Decimal('1.3')
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_decimal(benchmark, validate):
    serializer = Serializer(Decimal)
    benchmark.group = 'decimal'
    data = '1.3'
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


class FooEunm(enum.Enum):
    foo = 'foo'
    bar = 'bar'


def test_dump_enum(benchmark):
    serializer = Serializer(FooEunm)
    benchmark.group = 'enum'
    data = FooEunm.bar
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_enum(benchmark, validate):
    serializer = Serializer(FooEunm)
    benchmark.group = 'enum'
    data = 'foo'
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


@dataclass
class FooDataclass:
    foo: int
    bar: str


def test_dump_dataclass(benchmark):
    serializer = Serializer(FooDataclass)
    benchmark.group = 'dataclass'
    data = FooDataclass(foo=1, bar='2')
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_dataclass(benchmark, validate):
    serializer = Serializer(FooDataclass)
    benchmark.group = 'dataclass'
    data = {'foo': 1, 'bar': '2'}
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))


@dataclass
class Node:
    value: str
    next: Optional['Node'] = None


@dataclass
class Root:
    head: Node


def test_dump_recursive(benchmark):
    serializer = Serializer(Root)
    benchmark.group = 'recursive'
    data = Root(
        head=Node(
            value='1',
            next=Node(value='2'),
        ),
    )
    benchmark(repeat(lambda: serializer.dump(data)))


def test_load_recursive(benchmark, validate):
    serializer = Serializer(Root)
    benchmark.group = 'recursive'
    data = {'head': {'next': {'next': None, 'value': '2'}, 'value': '1'}}
    benchmark(repeat(lambda: serializer.load(data, validate=validate)))



from typing import Any

import marshmallow_dataclass

from .base import Dataclass, test_object


_schema = marshmallow_dataclass.class_schema(Dataclass)()

test_object = test_object


def load(data: dict[str, Any], validate: bool = True) -> Dataclass:
    return _schema.load(data)


def load_json(data: str, validate: bool = True) -> Dataclass:
    # todo: use orjson as render_module
    return _schema.loads(data)


def dump(obj: Dataclass) -> dict[str, Any]:
    return _schema.dump(obj)



import orjson
import pytest

from .compare.libs import serpyco_rs as serializer
from .utils import repeat


def test_dump(benchmark):
    serializer.dump(serializer.test_object)  # warmup
    benchmark.group = 'dump full'
    benchmark(repeat(lambda: serializer.dump(serializer.test_object), count=100))


def test_load(benchmark):
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=False)  # warmup

    benchmark.group = 'load full'
    benchmark(repeat(lambda: serializer.load(test_dict, validate=False), count=100))


def test_load_validate(benchmark):
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=True)  # warmup

    benchmark.group = 'load full with validate'
    benchmark(repeat(lambda: serializer.load(test_dict, validate=True), count=100))


@pytest.mark.parametrize('raw', [True, False], ids=['raw', 'python'])
def test_load_validate_compare(benchmark, raw):
    raw_data = orjson.dumps(serializer.dump(serializer.test_object)).decode('utf-8')

    serializer.load(orjson.loads(raw_data), validate=True)  # warmup
    serializer.load_json(raw_data, validate=True)  # warmup

    def callback():
        return serializer.load(orjson.loads(raw_data), validate=True)

    def raw_callback():
        return serializer.load_json(raw_data, validate=True)

    benchmark.group = 'load with validate (compare)'
    benchmark(repeat(raw_callback if raw else callback, count=100))



import json

import pytest

from .libs import marshmallow, mashumaro, pydantic, serpyco, serpyco_rs


serializers = {
    'serpyco_rs': serpyco_rs,
    'serpyco': serpyco,
    'pydantic': pydantic,
    'marshmallow': marshmallow,
    'mashumaro': mashumaro,
}


@pytest.mark.parametrize('lib', serializers.keys())
def test_dump(benchmark, lib):
    serializer = serializers[lib]
    serializer.dump(serializer.test_object)  # warmup

    benchmark.group = 'dump'
    benchmark.extra_info['lib'] = lib
    benchmark.extra_info['correct'] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    benchmark(serializer.dump, serializer.test_object)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load(benchmark, lib):
    serializer = serializers[lib]
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=False)  # warmup

    benchmark.group = 'load'
    benchmark.extra_info['lib'] = lib
    benchmark.extra_info['correct'] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    benchmark(serializer.load, test_dict, validate=False)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load_validate(benchmark, lib):
    serializer = serializers[lib]
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=True)  # warmup

    benchmark.group = 'load with validate'
    benchmark.extra_info['lib'] = lib
    benchmark.extra_info['correct'] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    benchmark(serializer.load, test_dict, validate=True)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load_json(benchmark, lib):
    serializer = serializers[lib]
    test_data = json.dumps(serializer.dump(serializer.test_object))
    serializer.load_json(test_data, validate=False)  # warmup

    benchmark.group = 'load raw json'
    benchmark.extra_info['lib'] = lib
    benchmark.extra_info['correct'] = serializer.load_json(test_data) == serializer.test_object
    benchmark(serializer.load_json, test_data, validate=False)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load_json_validate(benchmark, lib):
    serializer = serializers[lib]
    test_data = json.dumps(serializer.dump(serializer.test_object))
    serializer.load_json(test_data, validate=True)  # warmup

    benchmark.group = 'load raw json with validate'
    benchmark.extra_info['lib'] = lib
    benchmark.extra_info['correct'] = serializer.load_json(test_data) == serializer.test_object
    benchmark(serializer.load_json, test_data, validate=True)



import pytest
from serpyco_rs import SchemaValidationError, Serializer, ValidationError


def test_load_invalid_json__expected_validation_error():
    ser = Serializer(str)
    with pytest.raises(ValidationError) as exc_info:
        ser.load_json('{"a', validate=True)

    assert exc_info.value.args[0] == 'Invalid JSON string: EOF while parsing a string at line 1 column 3'


def test_load_json_invalid_type__expected_validation_error():
    ser = Serializer(str)
    with pytest.raises(SchemaValidationError) as exc_info:
        ser.load_json('123', validate=True)

    assert exc_info.value.errors[0].message == '123 is not of type "string"'



from typing import Any, Generic, TypeVar

from ._describe import Type

_T = TypeVar('_T')

class ValidationError(Exception):
    message: str

class ErrorItem:
    message: str
    instance_path: str
    schema_path: str

    def __init__(self, message: str, instance_path: str, schema_path: str): ...

class SchemaValidationError(ValidationError):
    errors: list[ErrorItem]

class Serializer(Generic[_T]):
    def __init__(self, py_class: Type, schema: str, pass_through_bytes: bool): ...
    def dump(self, value: _T) -> Any: ...
    def load(self, data: Any, validate: bool) -> _T: ...
    def load_json(self, data: str, validate: bool) -> _T: ...



import gc
import sys
from dataclasses import dataclass
from enum import Enum
from decimal import Decimal

import serpyco_rs


def test_repr_ref_count_segfault_on_default_value():
    class Bar(Enum):
        a = 'a'

    @dataclass
    class Foo:
        lang: Bar = Bar.a

    serializer = serpyco_rs.Serializer(Foo)

    for i in range(10):
        serializer.load({})
        serializer.load_json('{}')
        gc.collect()


def test_load_rc_count_decimal():
    serializer = serpyco_rs.Serializer(Decimal)

    val = '123.1'

    ref_count = sys.getrefcount(val)

    for i in range(100):
        serializer.load(val)
        serializer.load_json(f'"{val}"')
        gc.collect()

    assert ref_count == sys.getrefcount(val)


def test_dump_rc_count_decimal():
    serializer = serpyco_rs.Serializer(Decimal)

    val = Decimal('123.1')

    ref_count = sys.getrefcount(val)

    for i in range(100):
        serializer.dump(val)
        gc.collect()

    assert ref_count == sys.getrefcount(val)



import json
from typing import Annotated, Any, Generic, TypeVar, cast

from ._describe import describe_type
from ._impl import Serializer as _Serializer
from ._json_schema import get_json_schema
from .metadata import CamelCase, ForceDefaultForOptional, OmitNone


_T = TypeVar('_T', bound=Any)


class Serializer(Generic[_T]):
    def __init__(
        self,
        t: type[_T],
        *,
        camelcase_fields: bool = False,
        omit_none: bool = False,
        force_default_for_optional: bool = False,
        pass_through_bytes: bool = False,
    ) -> None:
        if camelcase_fields:
            t = cast(type[_T], Annotated[t, CamelCase])
        if omit_none:
            t = cast(type(_T), Annotated[t, OmitNone])  # type: ignore
        if force_default_for_optional:
            t = cast(type(_T), Annotated[t, ForceDefaultForOptional])  # type: ignore
        type_info = describe_type(t)
        self._schema = get_json_schema(type_info)
        self._encoder: _Serializer[_T] = _Serializer(
            type_info, schema=json.dumps(self._schema), pass_through_bytes=pass_through_bytes
        )

    def dump(self, value: _T) -> Any:
        return self._encoder.dump(value)

    def load(self, data: Any, validate: bool = True) -> _T:
        return self._encoder.load(data, validate=validate)

    def load_json(self, data: str, validate: bool = True) -> _T:
        return self._encoder.load_json(data, validate)

    def get_json_schema(self) -> dict[str, Any]:
        return self._schema



import os

import nox


nox.options.sessions = ['test', 'lint', 'type_check', 'rust_lint']
nox.options.python = False


def build(session):
    if _is_ci():
        # Install form wheels
        install(session, '--no-index', '--no-deps', '--find-links', 'wheels/', 'serpyco-rs')
        install(session, '--find-links', 'wheels/', 'serpyco-rs')
        return

    session.run_always('maturin', 'develop', '-r')

@nox.session(python=False)
def test(session):
    build(session)
    install(session, '-r', 'requirements/dev.txt')
    session.run('pytest', '-vss', 'tests/', *session.posargs)


@nox.session(python=False)
def lint(session):
    build(session)
    install(session, '-r', 'requirements/lint.txt')

    session.cd('python/serpyco_rs')
    paths = ['.', '../../tests', '../../bench']
    session.run('black', *(['--check', '--diff', *paths] if _is_ci() else paths))
    session.run('ruff', '.')


@nox.session(python=False)
def rust_lint(session):
    session.run('cargo', 'fmt', '--all', *(['--', '--check'] if _is_ci() else []))
    session.run('cargo', 'clippy', '--all-targets', '--all-features', '--', '-D', 'warnings')


@nox.session(python=False)
def type_check(session):
    build(session)
    install(session, '-r', 'requirements/type_check.txt')

    session.cd('python/serpyco_rs')
    session.run('pyright')
    session.run('pyright', '--verifytypes', 'serpyco_rs')
    session.run('mypy', '.', '--strict', '--implicit-reexport', '--pretty')


@nox.session(python=False)
def bench(session):
    build(session)
    install(session, '-r', 'requirements/bench.txt')

    session.run(
        'pytest',
        '--verbose',
        '--benchmark-min-time=0.5',
        '--benchmark-max-time=1',
        '--benchmark-disable-gc',
        '--benchmark-autosave',
        '--benchmark-save-data',
        '--benchmark-compare',
        *(session.posargs if session.posargs else ['bench']),
    )


@nox.session
def bench_codespeed(session):
    build(session)
    install(session, '-r', 'requirements/bench.txt')
    install(session, 'pytest-codspeed')
    session.run('pytest', 'bench', '--ignore=bench/compare', '--codspeed')


def _is_ci() -> bool:
    return bool(os.environ.get('CI', None))

def install(session, *args):
    session.run('pip', 'install', *args)
